# object.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from rechter.utils.file import list_files
from rechter.utils.time import fn_time, rtime
from rechter.utils.name import typed, name

import threading
import hashlib
import logging
import errno
import fcntl
import types
import json
import time
import os

## OBJECT

class Object(dict):

    """ basic Object on which the rest of the program is based. """

    def __getattribute__(zelf, name):
        if name in zelf: return zelf[name]
        return super().__getattribute__(name)

    def __getattr__(zelf, name):
        if name == "url": return zelf.urled()
        if name == "_ready":  zelf._ready = threading.Event()
        if name == "skip": return ["parsed", ] 
        if name == "cc": return "!"
        if name == "time": return zelf.timed()
        if name == "type": return typed(zelf)
        if name not in zelf: raise AttributeError(name)
        return zelf[name]

    def __contains__(zelf, name):    
        try: zelf[name] ; return True
        except KeyError: return False

    def __setattr__(zelf, name, value):
        if name in zelf and type(zelf[name]) in [types.FunctionType, types.MethodType]: raise WrongType(name)
        return dict.__setitem__(zelf, name, value)

    ## OUTPUT

    def reply(zelf, *args, **kwargs):
        """
            Send a reply to orginiating channel.

            args[0] = "channel to announce to"
            args[1] = "text to announce"

            >>> import rechter.bots
            >>> bot = rechter.bots.Bot()
            >>> e = bot.event("test")
            >>> e.reply("hello")
            hello
                   
        """
        if "outer" in zelf: zelf.outer.write(str(args[0]) + "\n") ; zelf.outer.flush() ; return
        if "channel" in zelf: zelf.say(zelf.channel, args[0])
        else: zelf.say(zelf.origin, args[0])

    def say(zelf, *args, **kwargs):
        """
            say text on a channel. use the _target attribute to relay the say request to. 

            args[0] = "channel to announce to"
            args[1] = "text to announce"

            >>> import rechter.bots
            >>> bot = rechter.bots.Bot()
            >>> e = bot.event("test")
            >>> e.say("#testchannel", "hello")
            hello

        """
        zelf._target.say(*args, **kwargs)

    def ok(zelf, *args, **kwargs):
        """
            reply with 'ok' and some descriptive text if provided. 

            args[0] = "optionale text to reply with"

            >>> import rechter.bots
            >>> bot = rechter.bots.Bot()
            >>> e = bot.event("test")
            >>> e.ok("done")
            ok done
                   
        """
        if "_target" in zelf: zelf.reply("ok %s" % " ".join([str(a) for a in args]))

    def show(zelf, *args, **kwargs):
        """
            return the data of the Object in key=value pairs

            >>> o = Object()
            >>> o.bla = "mekker"
            >>> o.show()
            ['bla=mekker']
                   
        """
        return ["%s=%s" % (a, zelf[a]) for a in sorted(zelf.names())]

    def display(zelf, *args, **kwargs):
        """
            Return a string constructed based on a list of keys. Include days ago as well.

            >>> import rechter.bots
            >>> bot = rechter.bots.Bot()
            >>> e = bot.event("test")
            >>> e.display(["txt", ])
            'test - '
                   
        """
        keys = args[0]
        txt = " ".join([str(getattr(zelf, str(key), "")) for key in keys])
        txt = txt.rstrip() 
        txt += " - %s" % zelf.days()
        return txt 

    def announce(zelf, *args, **kwargs):
        """
            Announce a text on all channels of all bots.

            args[0] = "text to announce"

            >>> import rechter.bots
            >>> bot = rechter.bots.Bot()
            >>> bot.announce("hello")
            hello
                   
        """
        from rechter.main import fleet
        thrs = []
        for bot in fleet: thrs.append(zelf.put(bot.announce, *args, **kwargs))
        if thrs: zelf.collect(thrs)
        return thrs

    ## SELECTOR

    def all(zelf, *args, **kwargs):
        """
            Select all records based on a optional prefix.

            args[0] = "optional prexix found in pathname"

            >>> o = Object()
            >>> res = o.all("cfg", "irc")
            >>> type(res)
            <class 'list'>
                   
        """
        a = list_files(zelf.path(), *args, **kwargs)
        return a

    def selected(zelf, *args, **kwargs):
        """
            Select all records based on a parsed. User provided string gives items to select on.

            args[0] = "parsed object found in events passed to command"

            >>> from rechter.utils.parse import txt_parse
            >>> from rechter.bots import Bot
            >>> bot = Bot()
            >>> p = txt_parse(bot, "find From=.nl time=2012-09 Subject")
            >>> result = bot.selected(p)
            >>> type(result)
            <class 'list'>

        """
        parsed = args[0]
        objs = []
        uniqlist = []
        for fn in zelf.all(parsed.args[0], **parsed.wanted):
            try: obj = Object().load(fn)
            except: logging.warn("fail %s" % fn) ; continue
            if "deleted" in obj and obj.deleted: continue
            if not obj.selector(parsed.args): continue
            if not obj.check_wanted(parsed.wanted): continue
            if obj.check_notwanted(parsed.not_wanted): continue
            uniq = parsed.switch.get("uniq", "")
            if uniq:
                val = obj.get(uniq, "")
                if val in uniqlist: continue
                else: uniqlist.append(val)
            objs.append(obj)
        return objs
 
    def selector(zelf, *args, **kwargs):
        """ 
            see if this objects has the desired attributes. 

            args[0] = "list of keys to test for"

            >>> from rechter.bots import Bot
            >>> bot = Bot()
            >>> from rechter.utils.parse import txt_parse
            >>> p = txt_parse(bot, "find From=.nl time=2012-09 Subject")
            >>> bot.selector(p.args)
            False
                   
        """
        keys = args[0]
        if not keys: return False
        for key in keys:
            if key == "time": continue
            if not getattr(zelf, key, None): return False
        return True

    def check_wanted(zelf, *args, **kwargs):
        """ 
            see if this objects has the desired attributes with their desired values.

            args[0] = "dictionary containing the desired key/values"

            >>> from rechter.utils.parse import txt_parse
            >>> from rechter.bots import Bot
            >>> bot = Bot()
            >>> o = txt_parse(bot, "find From=.nl time=2012-09 Subject")
            >>> o.check_wanted(o.wanted)
            True
                   
        """
        want = args[0]
        go = True
        for key, value in want.items():
           if key == "format": continue
           if key not in zelf: continue
           if value.startswith("-"): continue
           if value not in str(zelf[key]): go = False ; break 
        return go

    def check_notwanted(zelf, *args, **kwargs):
        """ 
            see if this object doesn't have not desired key/values.

            args[0] = "dictionary containing the not desired key/values"

            >>> from rechter.utils.parse import txt_parse
            >>> from rechter.bots import Bot
            >>> bot = Bot()
            >>> p = txt_parse(bot, "find From=.nl time=2012-09 Subject")
            >>> bot.check_notwanted(p.not_wanted)
            False
                   
        """
        not_want = args[0]
        for key, value in not_want.items():
           if key == "format": continue
           if key not in zelf: continue
           if value in zelf[key]: return True
        return False

    def allowed(zelf, *args, **kwargs):
        """ 
            method to see if a use to allowed based of origin of that user.

            args[0] = "dictionairy containing the not desired key/values"

            >>> import rechter.bots
            >>> bot = rechter.bots.Bot()
            >>> bot.allowed("root@cli")
            False
                   
        """
        origin = args[0]
        for fn in zelf.all("user"):
            obj = Object().load(fn)
            if "user" in obj and origin == obj.user: return True
        return False

    ## FLEET

    def search(zelf, *args, **kwargs):
        """ 
            search all attribute names. 

            args[0] = "attribute name to search for"

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.data = "test"
            >>> object.search("da")
            ['data']

        """
        l = []
        for key in zelf.names():
            if args[0] in key: l.append(key)
        return l

    def represent(zelf, *args, **kwargs):
        """ 
            return a string representation of the object.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.test1 = "yooo"
            >>> object.represent()
            'yooo'

        """
        return ", ".join([str(getattr(zelf, x, None)) for x in zelf.names() if getattr(zelf, x, None)])

    ## STATE

    def register(zelf, *args, **kwargs):
        """ 
            register a command, function and store in a list if value already exists

            >>> from rechter.object import Object
            >>> object = Object()
            >>> def test(event): event.reply("yooo")
            >>> object.register("test", test)
            >>> object["test"][0].cmnd
            'test'

        """
        from rechter.utils.trace import get_plugname
        o = Object()
        o.cmnd = args[0]
        o.func = args[1]
        o.plugname = get_plugname()
        if o.cmnd not in zelf: zelf[o.cmnd] = []
        todo = []
        for obj in zelf[o.cmnd]:
            if obj.cmnd == o.cmnd and obj.plugname == o.plugname: todo.append(obj)
        for obj in todo: zelf[o.cmnd].remove(obj) ; logging.info("! remove %s" % name(obj))
        zelf[o.cmnd].append(o)
        logging.info("! %s/register %s" % (zelf.type, o.cmnd))

    ## GETTERS

    def find(zelf, *args, **kwargs):
        """ 
            find all object based on the last value of their dotted keynames.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.find("ls")
            []

        """
        rest = args[0]
        result = []
        for x in set(zelf.search(rest)):
            if x.split(".")[-1] != rest: continue
            obj = zelf.get(x)
            if type(obj) in [list, tuple]: result.extend(obj) ; continue
            result.append(obj)
        return result

    def days(zelf, *args, **kwargs):
        """ 
            get the number of days relative to the ding's creation time.

            >>> from rechter.object import Object
            >>> import time
            >>> object = Object()
            >>> object.Date = time.ctime(time.time())
            >>> object.days()
            ''

        """
        from rechter.utils.time import elapsed
        t1 = time.time()
        t2 = zelf.timed()
        if t2:
            time_diff = float(t1 - t2)
            return elapsed(time_diff)

    def dated(zelf, *args, **kwargs):
        """ 
            Retrieve the creation time of this object.

            >>> from rechter.object import Object
            >>> import time
            >>> object = Object()
            >>> object.added = 'Sun Aug 23 16:07:16 2015'
            >>> object.dated()
            'Sun Aug 23 16:07:16 2015'

        """
        val = ""
        if "Date" in zelf: val = zelf.Date
        elif "date" in zelf: val = zelf.date
        elif "published" in zelf: val = zelf.published
        elif "added" in zelf: val = zelf.added
        elif "saved" in zelf: val = zelf.saved
        elif "timed" in zelf: val = zelf.timed
        return val

    def timed(zelf, *args, **kwargs):
        """ 
            Retrieve the time stored in the path name of this object.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.added = 'Sun Aug 23 16:07:16 2015'
            >>> object.timed()
            1440338836.0

        """
        from rechter.utils.time import short_date, to_time
        t = short_date(zelf.dated())
        if t: t = to_time(t)
        if not t and "_path" in zelf: t = fn_time(zelf._path)
        if not t: t = fn_time(rtime())
        return t

    def names(zelf, *args, **kwargs):
        """ 
            provide a list of keys suitable for iteration, skipping keys starting with a underscore

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.bla = "mekker"
            >>> list(object.names())
            ['bla']

        """
        for key in zelf.keys():
            k = str(key)
            if k.startswith("_"): continue
            yield key

    def slice(zelf, *args, **kwargs):
        """ 
            Provide a sub directory based of a optional provided key list

            args[0] = "keys to copy into slice of this object."

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object["bla"] = 1
            >>> object["bla2"] = 2
            >>> object.slice(["bla"])
            {'bla': 1}

        """
        o = Object()
        try: arguments = args[0]
        except IndexError: arguments = zelf.names()
        for arg in arguments:
            try: o[arg] = zelf[arg]
            except KeyError: continue
        return o

    def urled(zelf, *args, **kwargs):
        """ 
            url of the Object's file so that it can be retrieved when API server is running.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object._path = "bla"
            >>> object.urled()
            'http://wakker:10102/bla'

        """
        from rechter.cfg import cfg
        if "_path" in zelf: fn = os.sep.join(zelf._path.split(os.sep)[-2:])
        else: fn = ""
        return "http://%s:%s/%s" % (cfg.rest.hostname, cfg.main.port, fn)

    def root(zelf, *args, **kwargs):
        """ 
            return the working directory where the object is stored

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.root()
            '/home/bart/.rechter'

        """
        from rechter.utils.file import check_permissions
        from rechter.cfg import cfg
        if "workdir" in zelf: root = zelf.workdir
        else: root = cfg.main.workdir
        path = os.path.abspath(root)
        check_permissions(path)
        return path

    def path(zelf, *args, **kwargs):
        """ 
            return the path to where the object is stored

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.prefix = "cfg"
            >>> object.path()
            '/home/bart/.rechter/cfg'

        """
        from rechter.utils.join import j
        root = zelf.root()
        if "prefix" in zelf: root = j(root, zelf.prefix)
        return os.path.abspath(root) 

    ## INPUT

    def load(zelf, *args, **kwargs):
        """ 
            load a JSON file into this Object.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.load("bla")
            {'_path': 'bla'}

        """
        if args: path = args[0]
        else: path = zelf._path
        ondisk = zelf.read(path)
        fromdisk = json.loads(ondisk)
        if "data" in fromdisk: zelf.update(fromdisk["data"])
        else: zelf.update(fromdisk)
        if "saved" in fromdisk: zelf.saved = fromdisk["saved"]
        zelf._path = path
        logging.debug("load %s" % path)
        return zelf

    def read(zelf, *args, **kwargs):
        """ 
            load a JSON file into this Object.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.read("bla")
            '{}'

        """
        path = args[0]
        try: f = open(path, "r")
        except IOError as ex:
            if ex.errno == errno.ENOENT: return "{}"
            raise
        res = ""
        for line in f:
            if not line.strip().startswith("#"): res += line
        if not res.strip(): return "{}"
        f.close()
        return res

    ## PERSISTENCE

    def prepare(zelf, *args, **kwargs):
        """ 
            create JSON ready to be saved to disk.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> s = object.prepare()
            >>> type(s)
            <class 'str'>

        """
        from rechter.utils.trace import get_plugname
        #path = args[0]
        todisk = Object()
        todisk.data = zelf.slice()
        todisk.saved_from = get_plugname(2)
        todisk.type = zelf.type
        todisk.saved = zelf.saved = time.ctime(time.time())
        todisk.signature = todisk.data.make_signature()
        try: result = todisk.json(indent=2, ensure_ascii=False, sort_keys=True)
        except TypeError: raise NoJSON()
        return result

    def save(zelf, *args, **kwargs):
        """ 
            save Object as JSON to disk. use a provided filename or use a created timestamp.

            args[0] = "optional path relative to working directory"

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.save("bla")
            'bla'

        """
        from rechter.utils.time import rtime
        from rechter.utils.join import j
        if not args: t = rtime()
        else: t = args[0]
        zelf.sync(j(zelf.path(), t))
        return t

    def sync(zelf, *args, **kwargs):
        """ 
            sync Object to disk using the current path or use a timestamp.

            args[0] = "optional path to sync to"

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.prefix = "test"
            >>> path = object.sync()
            >>> '/home/bart/.rechter/test/' in path
            True

        """
        from rechter.utils.other import headertxt
        from rechter.utils.time import rtime
        from rechter.utils.file import cdir
        from rechter.utils.join import j
        try: path = args[0]
        except IndexError:
            try: path = zelf._path
            except AttributeError: path = zelf._path = j(zelf.path(), rtime())
        logging.info("! sync %s" % path)
        d, fn = os.path.split(path)
        cdir(d)
        #todisk = zelf.prepare(path, **kwargs)
        todisk = zelf.prepare(**kwargs)
        datafile = open(os.path.abspath(path) + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % "%s characters" % len(todisk))
        datafile.write(todisk)
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        os.rename(path + ".tmp", path)
        return path


    ## ITERATORS

    def first(zelf, *args, **kwargs):
        """ 
            find first object that has provided prefix.

            args[0] = "prefix argument to match in fn of Object"
            args[1] = "optional value to match"

            >>> from rechter.object import Object
            >>> object = Object()
            >>> o = object.first("cfg")
            >>> type(o)
            <class 'rechter.object.Object'>

        """

        fns = zelf.all(args[0])
        for fn in fns:
            obj = Object().load(fn)
            if "deleted" in obj and obj.deleted: continue
            if args and len(args) > 1 and args[1] != obj.get(args[0], ""): continue
            return obj

    def last(zelf, *args, **kwargs):
        """ 
            find last object that has provided prefix.

            args[0] = "prefix argument to match in fn of Object"
            args[1] = "optional value to match"

            >>> from rechter.object import Object
            >>> object = Object()
            >>> o = object.last("cfg")
            >>> type(o)
            <class 'rechter.object.Object'>

        """
        fns = zelf.all(args[0])[::-1]
        res = None
        for fn in fns[::-1]:
            obj = Object().load(fn)
            if "deleted" in obj and obj.deleted: continue
            if args and len(args) > 1 and args[1] != obj.get(args[0], ""): continue
            res = obj
        return res

    ## WAITERS

    def ready(zelf):
        """ signal to ready state. """
        logging.debug("! %s/ready" % zelf.type)
        zelf._ready.set()

    def clear(zelf):
        """ clear the ready state. """
        logging.debug("! %s/clean" % zelf.type)
        zelf._ready.clear()

    def wait(zelf, sec=180.0):
        """ wait for ready state. """
        logging.debug("! %s/wait" % zelf.type)
        zelf._ready.wait(sec)

    ## HELPERS

    def json(zelf, *args, **kwargs):
        """ 
            create JSON for this Object

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.bla = "mekker"
            >>> object.json()
            '{"bla": "mekker"}'

        """
        from rechter.utils.serialize import smooth
        return json.dumps(zelf.slice(*args, **kwargs), default=smooth, *args, **kwargs)

    def make_signature(zelf):
        """ 
            create a signature of this Object.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.make_signature()
            'bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f'

        """
        return str(hashlib.sha1(bytes(str(zelf), "utf-8")).hexdigest())

    def pretty(zelf, *args, **kwargs):
        """ 
            Make a pretty JSON string of this object.

            >>> from rechter.object import Object
            >>> object = Object()
            >>> object.bla = "mekker"
            >>> s = object.pretty()
            >>> type(s)
            <class 'str'>

        """
        from rechter.utils.serialize import pretty
        return json.dumps(zelf.slice(), indent=2, default=pretty, sort_keys=True, ensure_ascii=True)
